
from unittest import TestCase

from schema_checker.extras import pos_validator, kw_validator, args_validator


class TestExtras(TestCase):

    def test_pos_validator_ok(self):
        func = pos_validator(
            {
                'type': tuple,
                'value': int,
            }
        )(lambda *a, **b: (a, b))
        a, b = func(1, 2, 3, a='123')
        self.assertEqual(a, (1, 2, 3))
        self.assertEqual(b, {'a': '123'})

    def test_pos_validator_fail(self):
        func = pos_validator(
            {
                'type': tuple,
                'value': int,
            }
        )(lambda *a, **b: print(a, b) or (a, b))

        try:
            self.assertEqual(func(1, 2, '3', a='123'), ...)
        except ValueError:
            ...  # ok

    def test_kw_validator_ok(self):
        func = kw_validator(
            {
                'type': dict,
                'any_key': int,
            }
        )(lambda *a, **b: (a, b))
        a, b = func(1, '2', 3, a=123, b=42)
        self.assertEqual(a, (1, '2', 3))
        self.assertEqual(b, {'a': 123, 'b': 42})

    def test_kw_validator_fail(self):
        func = kw_validator(
            {
                'type': dict,
                'any_key': int,
            }
        )(lambda *a, **b: print(a, b) or (a, b))

        try:
            self.assertEqual(func(1, '2', 3, a='123', b=42), ...)
        except ValueError:
            ...  # ok

    def test_args_validator_ok(self):
        func = args_validator(
            pos_schema={
                'type': tuple,
                'value': int,
            },
            kw_schema={
                'type': dict,
                'any_key': int,
            },
        )(lambda *a, **b: (a, b))
        a, b = func(1, 2, 3, a=123, b=42)
        self.assertEqual(a, (1, 2, 3))
        self.assertEqual(b, {'a': 123, 'b': 42})

    def test_args_validator_fail_pos(self):
        func = args_validator(
            pos_schema={
                'type': tuple,
                'value': int,
            },
            kw_schema={
                'type': dict,
                'any_key': int,
            }
        )(lambda *a, **b: print(a, b) or (a, b))

        try:
            self.assertEqual(func(1, '2', 3, a=123, b=42), ...)
        except ValueError:
            ...  # ok

    def test_args_validator_fail_kw(self):
        func = args_validator(
            pos_schema={
                'type': tuple,
                'value': int,
            },
            kw_schema={
                'type': dict,
                'any_key': int,
            }
        )(lambda *a, **b: print(a, b) or (a, b))

        try:
            self.assertEqual(func(1, 2, 3, a='123', b=42), ...)
        except ValueError:
            ...  # ok


    def test_args_validator_fail_both(self):
        func = args_validator(
            pos_schema={
                'type': tuple,
                'value': int,
            },
            kw_schema={
                'type': dict,
                'any_key': int,
            }
        )(lambda *a, **b: print(a, b) or (a, b))

        try:
            self.assertEqual(func(1, '2', 3, a='123', b=42), ...)
        except ValueError:
            ...  # ok




from typing import Callable, Dict, Any, Type, Union
import functools

from .jschema import validate

SchemaType = Dict[str, Any]


def decorator_constructor(getter: Callable, setter: Callable):
    def validator(schema: Dict[str, Any]):
        def decorator(func):
            @functools.wraps(func)
            def wrap(*a, **b):
                a, b = setter(validate(getter(*a, **b), schema), a, b)
                return func(*a, **b)
            return wrap
        return decorator
    return validator


# ignores positional args
kw_validator = decorator_constructor(
    getter=lambda *a, **b: b,
    setter=lambda data, a, b: (a, data),
)

# ignores kw args
pos_validator = decorator_constructor(
    getter=lambda *a, **b: a,
    setter=lambda data, a, b: (data, b),
)


# validate both pos and kw args
def args_validator(pos_schema: SchemaType, kw_schema: SchemaType):
    def decorator(func):
        @functools.wraps(func)
        def wrap(*a, **b):
            a = validate(a, pos_schema)
            b = validate(b, kw_schema)
            return func(*a, **b)
        return wrap
    return decorator




import unittest

from schema_checker import validate


class TestJschema(unittest.TestCase):

    def do_test(self, obj, schema, result, expect=True, msg=None):
        try:
            if expect:
                self.assertEqual(validate(obj, schema), result)
            else:
                self.assertNotEqual(validate(obj, schema), result)
        except ValueError as e:
            self.assertFalse(expect)
            if msg:
                self.assertEqual(str(e), msg)

    def test_schema_fail(self):
        self.do_test(123, '123', 123, False)

    def test_simple_const(self):
        obj = '123'
        self.do_test(obj, 'const', obj)
        obj = 123
        self.do_test(obj, 'const', obj)

    def test_simple_str(self):
        obj = '123'
        self.do_test(obj, {'type': str}, obj)
        obj = 123
        self.do_test(obj, {'type': str}, obj, False)

    def test_simpler_str(self):
        obj = '123'
        self.do_test(obj, str, obj)
        obj = 123
        self.do_test(obj, str, obj, False)

    def test_simple_int(self):
        obj = 100500
        self.do_test(obj, {'type': int}, obj)
        obj = '100500'
        self.do_test(obj, {'type': int}, obj, False)

    def test_simpler_int(self):
        obj = 100500
        self.do_test(obj, int, obj)
        obj = '100500'
        self.do_test(obj, int, obj, False)

    def test_simple_float(self):
        obj = 1.001
        self.do_test(obj, {'type': float}, obj)
        obj = 1
        self.do_test(obj, {'type': float}, obj, False)

    def test_simpler_float(self):
        obj = 1.001
        self.do_test(obj, float, obj)
        obj = 1
        self.do_test(obj, float, obj, False)

    def test_simple_bool(self):
        obj = True
        self.do_test(obj, {'type': bool}, obj)
        obj = 1
        self.do_test(obj, {'type': bool}, obj, False)

    def test_simpler_bool(self):
        obj = True
        self.do_test(obj, bool, obj)
        obj = 1
        self.do_test(obj, bool, obj, False)

    def test_simple_dict(self):
        obj = {'1': 2, 'abc': 'cde'}
        self.do_test(
            obj,
            {
                'type': dict,
            },
            obj,
        )

    def test_dict(self):
        obj = {'1': 2, 'abc': 'cde'}
        self.do_test(
            obj,
            {
                'type': dict,
                'value': {
                    '1': int,
                    'abc': str,
                },
            },
            obj,
        )

    def test_dict_comples(self):
        obj = {
            '1': 123,
            '2': 42,
        }
        schema = {
            type: dict,
            'value': {
                '2': int,
                '3': {
                    type: int,
                    'default': 77,
                }
            },
            'unexpected': True,
        }
        self.do_test(
            obj=obj,
            schema=schema,
            result={
                '1': 123,
                '2': 42,
                '3': 77,
            }
        )

    def test_dict_fail(self):
        obj = {}
        self.do_test(
            obj,
            {
                'type': dict,
                'value': {
                    '1': {
                        'type': int,
                        'default': 2,
                    },
                    'abc': str,
                },
            },
            {'1': 2},
            False,
        )

    def test_simple_list(self):
        obj = ['1', 2, 'abc', 'cde']
        self.do_test(obj, {'type': list}, obj)

    def test_list(self):
        obj = ['1', '2', 'abc', 'cde']
        self.do_test(
            obj,
            {
                'type': list,
                'value': {
                    'type': str,
                }
            },
            obj,
        )

    def test_simple_tuple_ok(self):
        obj = ('1', 2, 'abc', 'cde')
        self.do_test(obj, {'type': tuple}, obj)

    def test_simple_tuple_fail(self):
        obj = ['1', 2, 'abc', 'cde']
        self.do_test(obj, {'type': tuple}, obj, False)

    def test_tuple_ok(self):
        obj = (1, 2, 123, 42)
        self.do_test(obj, {'type': tuple, 'value': int}, obj)

    def test_tuple_fail(self):
        obj = ('1', 2, 'abc', 'cde')
        self.do_test(obj, {'type': tuple, 'value': int}, obj, False)

    def test_simple_const_ok(self):
        obj = {'1': 123}
        schema = {
            'type': dict,
            'any_key': {
                'type': 'const',
                'value': 123,
            }
        }
        self.do_test(obj, schema, obj)

    def test_simple_const_fail(self):
        obj = {'1': 123}
        schema = {
            'type': dict,
            'any_key': {
                'type': 'const',
                'value': 122,
            }
        }
        self.do_test(obj, schema, obj, False)

    def test_const(self):
        obj = {
            '1': '2',
            '2': 10,
        }
        self.do_test(
            obj,
            {
                'type': dict,
                'value': {
                    '1': {
                        'type': 'const',
                        'value': '2',
                    },
                    '2': {
                        'type': 'const',
                        'value': 10,
                    },
                },
            },
            obj,
        )

    def test_default(self):
        obj = {'1': '2'}
        schema = {
            'type': dict,
            'value': {
                '1': {'type': str},
                'abc': {
                    'type': str,
                    'default': 'default_value',
                }
            }
        }
        res_obj = dict(obj, abc='default_value')
        self.do_test(obj, schema, res_obj)

    def test_filter(self):
        obj = {'1': 2}
        schema = {
            'type': dict,
            'value': {
                '1': {
                    'type': int,
                    'filter': lambda x: x < 5,
                },
            }
        }
        res_obj = obj
        self.do_test(obj, schema, res_obj)

    def test_filter_error(self):
        obj = {'1': 10}
        schema = {
            'type': dict,
            'value': {
                '1': {
                    'type': int,
                    'filter': lambda x: x < 5,
                },
            }
        }
        res_obj = obj
        self.do_test(obj, schema, res_obj, False)

    def test_pre_call(self):
        obj = {'1': '1, 2, 3, 4, 5, 6, 7, 8, 9, 10'}
        schema = {
            'type': dict,
            'value': {
                '1': {
                    'type': int,
                    'pre_call': lambda x: sum(int(i.strip()) for i in x.split(',')),
                },
            }
        }
        res_obj = {'1': sum(range(11))}
        self.do_test(obj, schema, res_obj)

    def test_post_call(self):
        obj = {'1': '2'}
        schema = {
            'type': dict,
            'value': {
                '1': {
                    # type will be checked after pre_call and before post_call
                    'pre_call': int,
                    'type': int,
                    'filter': lambda x: x < 5,
                    'post_call': str,
                },
            }
        }
        res_obj = {'1': '2'}
        self.do_test(obj, schema, res_obj)

    def test_blank(self):
        schema = {
            'type': str,
            'blank': True,
        }
        self.do_test('', schema, '', True)
        schema = {
            'type': str,
            'blank': False,
        }
        self.do_test('', schema, '', False)

    def test_max_length(self):
        schema = {
            'type': str,
            'max_length': 10,
        }
        self.do_test('A' * 10, schema, 'A' * 10, True)
        schema = {
            'type': str,
            'max_length': 10,
        }
        self.do_test('A' * 11, schema, 'A' * 11, False)

    def test_min_length(self):
        schema = {
            'type': str,
            'min_length': 10,
        }
        self.do_test('A' * 10, schema, 'A' * 10, True)
        schema = {
            'type': str,
            'min_length': 10,
        }
        self.do_test('A' * 9, schema, 'A' * 9, False)

    def test_unexpected_ok(self):
        schema = {
            'type': dict,
            'value': {
                '1': {
                    'type': int,
                },
            },
            'unexpected': True,
        }
        obj = {
            '1': 2,
            'abc': 'def',
        }
        self.do_test(obj, schema, obj, True)

    def test_unexpected_error(self):
        schema = {
            'type': dict,
            'value': {
                '1': {
                    'type': int,
                },
            },
            'unexpected': False,
        }
        obj = {
            '1': 2,
            'abc': 'def',
        }
        self.do_test(obj, schema, obj, False)

    def test_anykey_ok(self):
        schema = {
            'type': dict,
            'anykey': {
                'type': int,
            },
            'unexpected': True,
        }
        obj = {
            '1': 2,
            'abc': 3,
        }
        self.do_test(obj, schema, obj, True)

    def test_anykey_error(self):
        schema = {
            'type': dict,
            'any_key': {
                'type': int,
            },
            'unexpected': True,
        }
        obj = {
            '1': 2,
            'abc': '3',
        }
        self.do_test(obj, schema, obj, False)

    def test_type_union(self):
        schema = {
            'type': dict,
            'value': {
                'a': (int, float),
                'b': (str, bytes),
            }
        }
        obj = {'a': 1, 'b': '123'}
        self.do_test(obj, schema, obj)
        obj = {'a': 1.0, 'b': '123'}
        self.do_test(obj, schema, obj)
        obj = {'a': 1, 'b': b'123'}
        self.do_test(obj, schema, obj)
        obj = {'a': 1.0, 'b': b'123'}
        self.do_test(obj, schema, obj)

    def test_msg(self):
        schema = {
            'type': dict,
            'any_key': int,
            'errmsg': 'here must be int',
        }
        obj = {'123': 'not int'}
        self.do_test(obj, schema, obj, False, 'here must be int')

    def test_simple_enum(self):
        schema = {
            'type': 'enum',
            'value': {'1', '2'},
        }
        obj = '1'
        self.do_test(obj, schema, obj)
        obj = '2'
        self.do_test(obj, schema, obj)
        obj = '3'
        self.do_test(obj, schema, obj, False)

    def test_simple_enum_schema_fail(self):
        schema = {
            'type': 'enum',
        }
        obj = '1'
        self.do_test(obj, schema, obj, False)

    def test_enum(self):
        obj = {
            '1': '2',
            '2': 10,
        }
        self.do_test(
            obj,
            {
                'type': dict,
                'value': {
                    '1': {
                        'type': 'enum',
                        'value': {'1', '2'},
                    },
                    '2': {
                        'type': 'enum',
                        'value': range(100),
                    },
                }
            },
            obj,
        )



from typing import Any, Dict, NoReturn, TypeVar, Union, Type, Tuple

ObjType = TypeVar('ObjType')
SchemaType = Union[str, Type, Tuple[Type], Dict[Union[str, Type], Any]]


def _get_type(sch: Dict[Union[str, Type], Any]) -> Any:
    return sch[type if type in sch else 'type']


def _default(value: Any) -> Any:
    return value() if callable(value) else value


def _on_error(schema: SchemaType, msg: Union[str, Exception]) -> NoReturn:
    if isinstance(schema, dict):
        msg = schema.get('errmsg', msg)
    raise ValueError(msg)


def _validate_const_enum(obj: ObjType, schema: Dict[str, Any], schema_type: str, key: str) -> ObjType:
    if 'value' not in schema:
        _on_error(schema, 'schema for "enum" must contain "value"')
    if schema_type == 'enum':
        if obj not in schema['value']:
            _on_error(schema, '"{}" is not in enum "{}"')
    elif obj != schema['value']:
        _on_error(schema, '"{}" is not allowed as "{}"'.format(obj, key))
    return obj


def _check_dict_key(obj: ObjType, schema: Dict[str, Any], extra: str) -> ObjType:
    unex = {i for i in obj if i not in schema['value']}
    if unex and not schema.get('unexpected', False):
        _on_error(schema, 'Got unexpected keys: "{}" {};'.format('", "'.join([str(i) for i in unex]), extra))
    missed = {
        i
        for i in schema['value']
        if i not in obj and (not isinstance(schema['value'][i], dict) or 'default' not in schema['value'][i])
    }
    if missed:
        _on_error(schema, 'expected keys "{}" {}'.format('", "'.join([str(i) for i in missed]), extra))
    return {
        i: obj[i]
        for i in unex
    }


def _validate_dicts_value(obj: ObjType, schema: Dict[str, Any], extra: str) -> ObjType:
    new_obj = _check_dict_key(obj=obj, schema=schema, extra=extra)
    try:
        new_obj.update(
            {
                i: (
                    _default(schema['value'][i]['default'])
                    if i not in obj else
                    _apply(obj=obj[i], schema=schema['value'][i], key=i)
                )
                for i in schema['value']
            }
        )
    except ValueError as ex:
        _on_error(schema, ex)
    return new_obj


def _validate_dict(obj: ObjType, schema: Dict[str, Any], extra: str) -> ObjType:
    if 'value' in schema:
        obj = _validate_dicts_value(obj=obj, schema=schema, extra=extra)
    elif 'any_key' in schema:
        try:
            obj = {i: _apply(obj[i], schema['any_key'], i) for i in obj}
        except ValueError as ex:
            _on_error(schema, ex)
    return obj


def _generic_checks(obj: ObjType, schema: SchemaType, schema_type: Type, extra: str, key: str) -> ObjType:
    if not isinstance(obj, schema_type):
        _on_error(schema, 'expected type "{}" {} ; got {}'.format(schema_type, extra, type(obj)))
    if 'filter' in schema and not schema['filter'](obj):
        _on_error(schema, '"{}" not passed filter'.format(key))
    if schema.get('blank') is False and not obj:
        _on_error(schema, '"{}" is blank'.format(key))
    if 'max_length' in schema and len(obj) > schema['max_length']:
        _on_error(schema, '"{}" > max_length'.format(key))
    if 'min_length' in schema and len(obj) < schema['min_length']:
        _on_error(schema, '"{}" < min_length'.format(key))
    return obj


def _validate_generic(obj: ObjType, schema: SchemaType, schema_type: Type, key: str, extra: str) -> ObjType:
    obj = _generic_checks(obj=obj, schema=schema, schema_type=schema_type, key=key, extra=extra)
    if issubclass(schema_type, (list, tuple)) and 'value' in schema:
        try:
            obj = schema_type(_apply(i, schema['value'], key=key) for i in obj)
        except ValueError as ex:
            _on_error(schema, ex)
    elif issubclass(schema_type, dict):
        obj = _validate_dict(obj=obj, schema=schema, extra=extra)
    return obj


def _validate(obj: ObjType, schema: SchemaType, key: str, extra: str) -> ObjType:
    schema_type = _get_type(schema)
    if schema_type in {'const', 'enum'}:
        return _validate_const_enum(obj=obj, schema=schema, schema_type=schema_type, key=key)
    return _validate_generic(obj=obj, schema=schema, schema_type=schema_type, extra=extra, key=key)


def _apply(obj: ObjType, schema: SchemaType, key: str) -> ObjType:
    extra = ''.join(['for ', key]) if key else ''
    if not isinstance(schema, (dict, type, tuple)) and schema not in {'const', 'enum'}:
        raise ValueError('schema must be type, dict, tuple or "const"/"enum" {}'.format(extra))

    if schema == 'const':
        return obj

    if isinstance(schema, (type, tuple)):
        if isinstance(obj, schema):
            return obj
        raise ValueError('"{}" is not type of "{}" {}'.format(obj, schema, extra))

    if 'pre_call' in schema:
        obj = schema['pre_call'](obj)

    obj = _validate(obj=obj, schema=schema, key=key, extra=extra)

    if 'post_call' in schema:
        obj = schema['post_call'](obj)
    return obj


def validate(obj: ObjType, schema: SchemaType) -> ObjType:
    """
        obj - some object
        schema - schema_checker
        schema ::= type of this object : list/dict/str/int/float (can be tuple of types) or "const"/"enum"
          OR
        schema ::= dict - {
          type         : type of this object : "list/tuple/dict/str/int/float or "const"
          "value"      : need for obj type of
                           - list/tuple - is schema for all elements in list
                           - dict - dict[key -> schema]
                           - const - some value to be compared with using method
                           - enum - list/set/dict/tuple to check if obj __contains__ in "value"
          "any_key"     : need for obj type of dict - schema for all keys (ignores if value is set)
          "default"    : default value if this object does not exists (if callable will be called)
          "filter"     : function value -> bool - if false then raise error
          "pre_call"   : function value -> value - will be called before cheking filter and value
          "post_call"  : function value -> value - will be called after cheking filter and value
          "blank"      : raise error if value is blank
          "max_length" : extra check of length (len)
          "min_length" : extra check of length (len)
          "unexpected" : allow unexpected keys (for dict)
          "errmsg"     : will be in ValueError in case of error on this level
        }
    """
    return _apply(obj, schema, 'Top-level')